/* 32-bit Forth in C for Win32.         Copyright  Robert. F. Illyes  1994-2000
-------------------------------------------------------------------------------
This FORTH may be copied for personal use at no cost, provided the copyright
notices are intact and this file is included unaltered. All other rights are
reserved. The sale and/or publication of this Forth is prohibited without the
written consent of Robert F. Illyes. No warranty is made regarding the fitness
of this FORTH for any particular purpose.

This Forth permits the easy adaptation of 16-bit graphics programs to Windows 95
or later. Many words work with both the 32-bit, 7-digit floating point and the
32-bit integers: stack words such as DUP, memory words such as !, and a few
other words including 0<, 0=, =, 0, VARIABLE, and CONSTANT. Numeric input is set
to integer or floating point by INT and FP. Floating point input may be forced
in INT mode by including a decimal point.

This is an emulation of a Forth CPU. It has no address interpreter, but
rather a subroutine call and return. Definitions after ":" in the dictionary are
in Forth and compile as the subroutine op code followed by an index into the ram
array. Lower definitions are of op codes and compile as a C procedure pointer. '
gives a definition's header index. AT converts the index to the dictionary
address if a high-level definition. | combines the actions of COMPILE and
[COMPILE]. It is assumed that pointers, floating point values, and integers on
the stack and in the ram array are all 32-bit. If this is being adapted to a
machine for which this is not the case, several unions will have to be added.
They were not used here because they make the code less readable. The code for
UM/MOD also assumes that a conversion

A Pentium CPU has 2 integer and 1 floating point pipelines, each capable of one
operation per clock cycle. Arithmetic proceeds in parallel with subroutine
handling, so the maximum execution speed of this Forth is determined by the time
required for a C subroutine call and return. Execution would be up to 5 times
faster if this emulation overhead were eliminated, and up to 10 times faster
with peephole optimization. There are several commercial Forths that achieve
this speed. If maximum speed is needed, it is important that the Pentium's
floating point stack be directly usable, but there is no speed advantage to this
approach with this Forth, and there is only one parameter stack.

This code should be compiled with a 32-bit C++ compiler with maximum Pentium
optimization. When more speed is needed, C routines should be added and made op
codes. For the graphics and image processing studies for which this Forth is
intended, speed is already completely satisfactory in most cases on the newer
Pentiums. Note that the emulation overhead becomes negligible with complex tasks
such as line drawing. */

#include <time.h>
#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include <windows.h>
#include <wingdi.h>
#include <winnt.h>

#define  PUSH(N)  *++s = t; t = N         /* Push N onto parameter stack      */
#define  POP      *s--                    /* Pop second stack value           */
#define  DROP     t=POP                   /* Forth drop in C                  */
#define  FPOP     *(float *)s--           /* POP in floating point            */
#define  SAID     heads[top+1].name       /* word() leaves word here          */
//
char *in,*xin,tib[256],fib[256];          /* I/O strings and pointers         */
typedef void(func)(void);
struct header { 			  /* Dictionary header structure      */
 union {func * op; int ram;} code;        /* Pointer to C or Forth procedure  */
 char name[26]; 			  /* Name of dictionary entry	      */
 char immediate;			  /* True if immediate definition     */
} *heads;
static union { float    f;                // Anonomous union containing the
               int      t;                // top of the parameter stack
               unsigned u;
               func    *p; } ;
int   stack[100],	/* The parameter stack. 			      */
      *s=stack, 	/* The parameter stack pointer. 		      */
      op,		/* Op code found by find()			      */
      pc,		/* The program counter (index to an element of ram[]. */
      state=0,		/* Non-zero during compilation. 		      */
      ram[100000],	/* Compiled definitions with numeric base initialized.*/
      rs[100],		/* The return stack, an array of indices into ram.    */
      *r=rs,		/* The return stack pointer.			      */
      *h=ram+3,         /* The pointer to HERE.                               */
      top=199,		/* The index of the last used header.		      */
      ctop;		/* The index of the last header of an op code.	      */
/*
Character output and related procedures
-------------------------------------------------------------------------------
*/
HWND	    window;	// Window handle
HDC	    bitmap;	// Bitmap device context
HBITMAP     bits;	// Bitmap handle
PAINTSTRUCT paint;	// Painting structure
HANDLE      input=0,    // Handle to INCLUDE or other input file
            output=0;   // Handle to output file
int         high = 433, // Window height
            wide = 649, // Window width
	    el	 =   0, // Line number for text output
	    x	 =   0, // X coordinate for next character
	    y	 =   0; // Y coordinate for next character
HPEN	    pen;	// Pen used for line drawing
COLORREF    color;	// Current drawing and plotting color
int	    newcolor=1; // True if there is a new color

void apply()	   { if (newcolor) {
			pen = CreatePen(PS_SOLID, 1, color);	 // Is this all
			DeleteObject(SelectObject(bitmap, pen)); // necessary?
			SetTextColor(bitmap, color);
                        newcolor=0; }}
void eset()        { y = t; x = POP; DROP; }
void emt(char c)   { TextOut(bitmap, x, y, &c, 1); }
void cr()          { x = 0; y = 12 * el++;             // SHOULD BE IN emit()!!
                     if (y > 400) { el--; y -= 12;
                        BitBlt(bitmap, 0,  0, wide, high-12,
                               bitmap, 0, 12, SRCCOPY); } }
void emit()        { if (output) ; // WriteFile(output, &(char)t, 1, &have, NULL);
                     else { apply(); emt(t); x += 8; if (x > (wide-10)) cr(); }
                     DROP; }
void sout(char *c) { while(*c) { PUSH(*c++); emit(); } }
void clear()       { PatBlt(bitmap, 0, 0, wide, high, WHITENESS); el=x=y=0; }

/*
Window callback function and related code
-------------------------------------------------------------------------------
In addition to handling program termination and window repainting, wfunc()
enters keystrokes into the text input buffer and calls the Forth interpreter
each time a newline is received.
*/
void interpret();
int  interpreting = 0, // True if interpreting
     keypress     = 0; // 1-character type-ahead buffer

void MessageCheck(){ MSG msg;
		     while(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
		       if (!GetMessage(&msg, NULL, 0, 0)) exit(msg.wParam);
                       TranslateMessage(&msg);
		       DispatchMessage(&msg); } }
void force()	   { InvalidateRect(window, NULL, 0); }
void repaint()     { force(); MessageCheck(); }
void key()         { do MessageCheck(); while(!keypress);
                     PUSH(keypress); keypress = 0; }
void qkey()        { MessageCheck(); PUSH(keypress); }

LRESULT CALLBACK wfunc(HWND window, UINT message, WPARAM w, LPARAM l) {
  switch(message) {
    case WM_CHAR:   if (w== 3) {  DestroyWindow(window); // Control-c (doesn't
                                  break; }               // work during interpret.
                    if (interpreting) { keypress = w;    // Prevents reentry of
                                  break; }               // query & interpret.
		    if (w==13) { *in = 0;
				  in = tib;
				  PUSH(' ');
				  emit();
                                  interpreting = 1;
				  interpret();
                                  interpreting = 0;
				  cr();
                                  in = tib;
                                  if (keypress)
                                    SendMessage(window, WM_CHAR, keypress, 0);
                                  keypress = 0; }
                     else { if (w==8) { if (in>tib) { emt(' ');
						      if (in>tib) in--;
						      x -= 8; } }
			    else      { *in = w;
					if ((in-tib) < 255) in++;
					PUSH(w); emit(); } }
                     emt('_');
		     force();
		     break;
    case WM_PAINT  : BitBlt(BeginPaint(window,&paint), 0, 0, wide, high,
                                       bitmap, 0, 0, SRCCOPY);
                     EndPaint(window, &paint);
		     break;
    case WM_DESTROY: DeleteObject(pen );
		     DeleteObject(bits);
		     PostQuitMessage(0);
		     break;
    default: return DefWindowProc(window, message, w, l); }
  return 0;
} /* 
File access for source and text
---------------------------------------------------------------------------
Note that the "(" comment goes only to the end of a text file record if there
is no paired ")". So the "//" comment is not technically necessary. But it
is nice because it allows the inclusion of right parentheses.

Text is placed in char fib[256] which is not directly available in Forth. But
it can be parsed by WORD, which leaves words and count at HERE.

record() and character() automatically close the file when end-of-file is
reached and return a false.
*/
int   get();
void  warn(char *);
DWORD have;	       // End of file indicator for record()
char  ch;	       // Character acquired by character()

int  opener(int n) { xin = in;  // Save input pointer
     char name[20]={0};
     int count = ram[n];
     for (int i=0; i<count; i++) name[i] = ram[i + n + 1]; // Build string
     input = CreateFile(name, GENERIC_READ, FILE_SHARE_READ,
                NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
     if (input == INVALID_HANDLE_VALUE) { input = 0; sout(" No "); sout(name); }
     return !input;
}
void openf()  { get(); PUSH(opener(h - ram)); }    // Filename is next word
void opens()  { t = opener(t); }                   // Open from Forth string
//void fil()    {
//void scr()

int  character()  {
     ReadFile(input, &ch, 1, &have, NULL);	   // Read one byte
     if (!have) {				   // If end of file,
       in = xin;				   // restore input pointer,
       CloseHandle(input);			   // close file, and
       input = NULL; }				   // indicate keyboard input.
     return have;
}
void chars() { PUSH(0); PUSH(character()); *s=t; }   // Forth adaptation

int  record() { in = fib;	  // Set input to file buffer.
     while(character())  {	  // Stop at end of file.
	*in = ch;		  // Get character and store.
	if (ch	== '\n') break;   // Stop at linefeed.
	if (*in != '\r') in++; }  // Ignor carriage returns.
     if (have){*in = 0; 	  // Replace newline with null.
		in = fib; }	  // Input pointer back to beginning
     return have;
}
void frecord() { PUSH(record());} // Forth version of record

/*
Image processing vocabulary, 2-byte signed pixels, 256 column
-------------------------------------------------------------------------------
*/
char* pix;          // Pointer to the bitmap buffer
int   bottom=255;   // Last line read by image()
short a[131072],    // Image buffers
      *anear;	    // Pointer into image buffer b

void image() { int i = 0, line = 0;               // Read an image from a file
     openf(); if (t) { DROP; return; }
     while(character()) if(i < 65536) a[i++] = ch & 255;
     bottom = i / 256;
}
void bmp() {
     openf(); if (t) DROP; return; }
void gplot(int x, int y, int g) {
     char * addr = pix + x*3 + 1948 * (432 - y);
     *addr++ = g; *addr++ = g; *addr = g;
}
void scrn()  { t = *(char *) (pix + POP*3 + 1948 * (432-t)); }
void vplot() { t = max(0, min(255,t)); int y = POP; gplot(POP, y, t); DROP; }
void see() {			       // Display image buffer a
     if (t >= *s && s-stack > 1) {
      int v=30, r, gray;
      double scale = 256. / (t - *s);
      for (int y=0; y<bottom; y++) {
	 r = 256 * y;
	 for (int x=0; x<256; x++) {
            gray = max(0, min(255, (int) ((a[x + r] - *s) * scale) ));
            gplot(x+355, v, gray);
            if (ram[1] && !(v%6)) gplot(x+355, v+1, gray); }
         v += 1 + (ram[1] && !(v%6)); }
      RECT rect = {355, 30, 611, 286};
      InvalidateRect(window, &rect, 0);
      MessageCheck(); }
     else sout("Invalid range. ");
     DROP; DROP;
}
void azero()  { for(int i=0; i<65536; i++) a[i]=0; bottom=255; }
void xch()    { for (int i=0; i<65536; i++) {
		  int p = a[i];
		  a[i] = a[i+65536];
		  a[i+65536] = p; } }
void nearto() { anear = a + 256 * t + POP; DROP; }
void afetch() { PUSH(*(anear + ram[pc++])); }
void astore() { *(anear + ram[pc++]) = t; DROP; }
/*
C procedures for most op codes
-------------------------------------------------------------------------------
To understand how a compiled Forth word is executed, take a close look at
execute(). More important procedures are grouped at the beginning to make it
likely that they are all in cache memory.

The use of a longjmp in execute() to exit the execution loop is marginally
faster than having all forth op codes return a zero except for kill() or by
testing for an empty return stack as was done in the original 1994 version of
this Forth.
*/
int find(char*), addr(int), wort(char);
void compile(), call(int), literal(), number(), comment();

jmp_buf environment;

void kill()   { longjmp(environment,1); }               // See use in ABORT
void execute(){ int n = t; DROP;                        // Hide top of stack
                if (n>ctop) {                           // If high-level:
                   pc = heads[n].code.ram;              //  Init prog counter.
                   *++r = addr(find("ABORT"));          //  Push ABORT.
                   if (setjmp(environment) == 0)        //  Save environment.
                      while(1)                          //  Execute Forth till
                         ((func *) ram[pc++])(); }
                else (heads[n].code.op)(); }            // Else Execute op code.

void nest()   { *++r = pc+1;  pc = ram[pc]; }
void unnest() { pc = *r--; }

void lit()    { PUSH(ram[pc++]); }
void zero()   { PUSH(0); }
void store()  { ram[t] = POP; DROP; }
void pstore() { ram[t]+= POP; DROP; }
void fetch()  { t = ram[t]; }

void ifs()    { pc = (t)?  pc+1:  ram[pc];  DROP; }
void branch() { pc = ram[pc]; }
void rfetch() { PUSH(*r    ); }
void j()      { PUSH(*(r-2)); }
void rpull()  { PUSH(*r--  ); }
void rput()   { *++r = t; DROP; }
void loop()   { if ( ++*r == *(r-1)) { ++pc; r-=2; }
		else   pc = ram[pc]; }
void ploop()  { if (((*r-*(r-1)) ^ (*r-*(r-1)+t)) < 0) { ++pc; r-=2; }
		else { pc = ram[pc]; *r += t; } DROP; }

void plus()   { t += POP; }
void minus()  { t  = POP-t; }
void negate() { t  = -t; }
void inc()    { t++; }
void dec()    { t--; }
void twice()  { t +=  t; }
void half()   { t >>= 1; }
//
void drop()   { DROP; }
void dup()    { *++s = t; }
void over()   { PUSH(*(s-1)); }
void swap()   { int i = t;  t = *s;  *s = i; }
void rot()    { int i = *(s-1); *(s-1) = *s; *s = t; t=i; }
void and()    { t &= POP; }

void lt()     { t  = -(POP <  t); }
void gt()     { t  = -(POP >  t); }
void eq()     { t  = -(POP == t); }
void not()    { t  = ~t;  }
void or()     { t |= POP; }
void xor()    { t ^= POP; }
void eqz()    { t  = -!t; }
void neg()    { t  = -(0   >  t); }
void ult()    { t  = -((unsigned) POP < u); }

void dup2()   { int a = *s; *++s = t; *++s = a; }
void ehalf()  { int i = t>>1; t &= i; t &= 1; t += i; }
void sabs()   { t  = abs(t); }
void smod()   { t  = POP%t; }
void mult()   { t *= POP; }
void iabs()   { t  = abs(t); }
void sdiv()   { t = POP/t;	}
void imax()   { t = max(t, POP); }
void imin()   { t = min(t, POP); }
void sq()     { t *= t; }

void fgt()    { f  = -(FPOP > f); }
void flet()   { f  = -(FPOP < f); }
void fmax()   { f  = max(f, FPOP); }
void fdot()   { char n[20]; sprintf(n,"%.7g ", *(float *)&t); DROP; sout(n); }
void edot()   { char n[20]; sprintf(n,"%.7e ", *(float *)&t); DROP; sout(n); }
void usqrt()  { t = sqrt(t); }
void flt()    { f  = t; }
void fplus()  { f += FPOP; }
void fmult()  { f *= FPOP; }
void fminus() { f  = FPOP-f; }
void fdiv()   { f  = FPOP/f; }
void fnegate(){ f  = -f;     }
void d_to_r() { f *= .017453293 ; }        // /90  .01745329251994329577;
void fsin()   { f  = sin(f); }
void fcos()   { f  = cos(f); }
void fsqrt()  { f  = sqrt(f);}
void fsq()    { f *= f;      }
void flen()   { double g = FPOP; f = sqrt(f*f + g*g); }
void fexp()   { f  = exp(f); }
void fabsv()  { f  = fabs(f);}
void ffix()   { t  = f;   }
void fround() { if (f<0) t=f-.5; else t=f+.5; }
void ftwice() { f += f; }
void fln()    { f  = log(f); }
void flog()   { f  = log10(f); }
void fpow()   { f  = pow(f, FPOP); }
void ftan()   { f  = tan(f); }
void fatan()  { f  = atan(f); }
//
void dinc()   { t += !++*s; }
void dtwice() { t = (t<<1) + (*s<0); *s <<= 1; }
void dnegate(){ *s = ~*s + 1;  t = ~t + !*s; }
void dplus()  { int l = POP; t += POP; int n = (l<0) + (*s<0); *s += l;
		if (n) { if (n==2) t++; else if ((*s)>=0) t++; } }
void ummul()  { union { int i[2]; DWORDLONG l; } db;
                db.l = UInt32x32To64(u, *(unsigned *) s);
                t = db.i[1]; *s = db.i[0]; }
void mmul()   { union { int i[2]; LONGLONG l; } db;
                db.l = Int32x32To64(t,*s); t = db.i[1]; *s = db.i[0]; }
void muldiv() { int m = POP; t = MulDiv(POP, m, t); }
void rdiv()   { t = MulDiv(1, POP, t); }
void umdiv()  { unsigned a = t, b = POP;
                long double d = ldexpl(b, 32) +(unsigned) *s;
                u = d / a;   *(unsigned *)s = d - a * u; }
void drop2()  { POP; DROP; }
void head()   { get(); top++; heads[top].code.ram=h-ram; heads[top].immediate=0; }
void comp()   { compile(); }
void colon()  { head(); compile(); }
void fill()   { int *p, i=POP; p=ram+POP; while (i--) *p++=t; DROP; }
void here()   { PUSH(h-ram); }
void allot()  { h += t;  DROP; }
void depth()  { PUSH(s-stack-1); }
void literal(){ *h++ = heads[2].code.ram; *h++ = t; DROP; }
void last()   { PUSH(top); }


void immed()  { heads[top].immediate = 1; }
void calls()  { call(t); DROP; }
void tick()   { get(); PUSH(find(SAID)); }
void forget() { tick(); if (t>ctop) { top = t-1; h = heads[t].code.ram + ram; }
		DROP; }
void bar()    { tick(); if (!heads[t].immediate) {literal(); PUSH(find("CALL"));}
		calls(); }
void at()     { t = addr(t); }
void states() { PUSH(state); }
void incl()   { openf(); if(t) record(); DROP;}
void word()   { wort(t); t = h - ram; }
void words()  { for (int i=1; i<=top; i++) { sout("  "); sout(heads[i].name); } }


void dots()   { int i; *(s+1)=t;
                for (i=1 ;i <=  s-stack; i++) printf("  %i", stack[i+1]);}


void bye()    { DestroyWindow(window); }
void forth()  { system("forth"); }
void plot()   { SetPixelV(bitmap, POP, t, color); DROP; }
void set()    { MoveToEx(bitmap, POP, t, NULL ); DROP; }
void line()   { apply();  LineTo(bitmap, POP, t); DROP; }
void rgb()    { int b=t, g=POP, r=POP;  color = RGB(r, g, b); newcolor=1; DROP;}
void gray()   { t = min(max(t,0),255); color = RGB(t, t, t); newcolor=1; DROP; }
void icos()   { float c = 10000 * cos(.017453293 * abs(t));
		if (c<0) t = (int) (c-.5); else t = (int) (c+.5); }
void isin()   { t -= 90; icos(); }
void ms()     { PUSH(clock()); }
/*
Create separate header structure and initialize for op codes
-----------------------------------------------------------------------------*/
struct header ops[1000] = {

/* Structures */   nest   ,  ""         ,  0 ,     unnest ,  ";"      ,  0 ,
		   lit	  ,  "lit"      ,  0 ,     ifs    ,  "if"     ,  0 ,
		   branch ,  "branch"   ,  0 ,     loop   ,  "loop"   ,  0 ,
		   ploop  ,  "+loop"    ,  0 ,     unnest ,  "EXIT"   ,  0 ,
/* Memory     */   fetch  ,  "@"        ,  0 ,     store  ,  "!"      ,  0 ,
		   fill   ,  "FILL"     ,  0 ,     pstore ,  "+!"     ,  0 ,
/* Arithmetic */   plus   ,  "+"        ,  0 ,     minus  ,  "-"      ,  0 ,
		   sdiv   ,  "/"        ,  0 ,     smod   ,  "MOD"    ,  0 ,
                   rdiv   ,  "/R"       ,  0 ,
		   mult   ,  "*"        ,  0 ,     twice  ,  "2*"     ,  0 ,
                   muldiv ,  "*/"       ,  0 ,     umdiv  ,  "UM/MOD" ,  0 ,
		   half   ,  "2/"       ,  0 ,     inc    ,  "1+"     ,  0 ,
		   dec	  ,  "1-"       ,  0 ,     negate ,  "NEGATE" ,  0 ,
                   iabs   ,  "ABS"      ,  0 ,     ummul  ,  "UM*"    ,  0 ,
		   usqrt  ,  "SQRT"     ,  0 ,     sabs   ,  "ABS"    ,  0 ,
		   flt	  ,  "FLOAT"    ,  0 ,     ffix   ,  "FIX"    ,  0 ,
		   fround ,  "ROUND"    ,  0 ,     ehalf  ,  "2/E"    ,  0 ,
		   fplus  ,  "F+"       ,  0 ,     fabsv  ,  "FABS"   ,  0 ,
		   fminus ,  "F-"       ,  0 ,     fmult  ,  "F*"     ,  0 ,
		   fdiv   ,  "F/"       ,  0 ,     fnegate,  "FNEGATE",  0 ,
		   fcos   ,  "FCOS"     ,  0 ,     fsin   ,  "FSIN"   ,  0 ,
		   fsqrt  ,  "FSQRT"    ,  0 ,     fsq    ,  "FSQ"    ,  0 ,
                   flen   ,  "FLEN"     ,  0 ,
                   fexp   ,  "FEXP"     ,  0 ,     ftan   ,  "FTAN"   ,  0 ,
                   fatan  ,  "FATAN"    ,  0 ,     fpow   ,  "F^"     ,  0 ,
                   flog   ,  "FLOG"     ,  0 ,     fln    ,  "FLN"    ,  0 ,
                   ftwice ,  "F2*"      ,  0 ,     mmul   ,  "M*"     ,  0 ,
		   dplus  ,  "D+"       ,  0 ,     dnegate,  "DNEGATE",  0 ,
		   dinc   ,  "D1+"      ,  0 ,     dtwice ,  "D2*"    ,  0 ,
                   imin   ,  "MIN"      ,  0 ,     imax   ,  "MAX"    ,  0 ,
                   sq     ,  "SQ"       ,  0 ,     d_to_r ,  "DEG>RAD" , 0 ,
/* Boolean    */   not	  ,  "NOT"      ,  0 ,     and    ,  "AND"    ,  0 ,
		   or	  ,  "OR"       ,  0 ,     xor    ,  "XOR"    ,  0 ,
/* Comparison */   lt	  ,  "<"        ,  0 ,     eqz    ,  "0="     ,  0 ,
		   neg	  ,  "0<"       ,  0 ,     eq     ,  "="      ,  0 ,
                   gt     ,  ">"        ,  0 ,     ult    ,  "U<"     ,  0 ,
		   flet   ,  "F<"       ,  0 ,     fgt    ,  "F>"     ,  0 ,
		   fmax   ,  "FMAX"     ,  0 ,
/* Stacks     */   drop   ,  "DROP"     ,  0 ,     dup    ,  "DUP"    ,  0 ,
		   over   ,  "OVER"     ,  0 ,     swap   ,  "SWAP"   ,  0 ,
		   rput   ,  ">R"       ,  0 ,     rpull  ,  "R>"     ,  0 ,
		   rfetch ,  "R@"       ,  0 ,     j      ,  "J"      ,  0 ,
		   rfetch ,  "I"        ,  0 ,     rot    ,  "ROT"    ,  0 ,
                   dup2   ,  "2DUP"     ,  0 ,     drop2  ,  "2DROP"  ,  0 ,
/* I/O	      */   fdot   ,  "F."       ,  0 ,     edot   ,  "E."     ,  0 ,
                   key    ,  "KEY"      ,  0 ,     emit   ,  "EMIT"   ,  0 ,
		   frecord,  "RECORD"   ,  0 ,     opens  ,  "OPEN"   ,  0 ,
		   eset   ,  "EMITSET"  ,  0 ,     cr     ,  "CR"     ,  0 ,
                   number ,  "NUMBER"   ,  0 ,     qkey   ,  "?KEY"   ,  0 ,
                   chars  ,  "CHAR"     ,  0 ,
//
/* Graphics   */   clear  ,  "CLEAR"    ,  0 ,     plot   ,  "PLOT"   ,  0 ,
		   set	  ,  "SET"      ,  0 ,     line   ,  "LINE"   ,  0 ,
		   rgb	  ,  "RGB"      ,  0 ,     repaint,  "REPAINT",  0 ,
		   isin   ,  "SIN"      ,  0 ,     icos   ,  "COS"    ,  0 ,
                   gray   ,  "GRAY"     ,  0 ,
                   vplot  ,  "VPLOT"    ,  0 ,     scrn   ,  "SCRN"   ,  0 ,
/* Images     */   see	  ,  "SEE"      ,  0 ,     image  ,  "IMAGE"  ,  0 ,
		   astore ,  "astore"   ,  0 ,     afetch ,  "afetch" ,  0 ,
                   nearto ,  "NEAR"     ,  0 ,     azero  ,  "MZERO"  ,  0 ,
                   xch    ,  "XCH"      ,  0 ,
/* Compiler   */   immed  ,  "IMMEDIATE",  0 ,     here   ,  "HERE"   ,  0 ,
		   allot  ,  "ALLOT"    ,  0 ,     head   ,  "HEAD"   ,  0 ,
		   last   ,  "LAST"     ,  0 ,     tick   ,  "'"      ,  0 ,
		   zero   ,  "0"        ,  0 ,     depth  ,  "DEPTH"  ,  0 ,
		   at	  ,  "AT"       ,  0 ,     bar    ,  "|"      ,  1 ,
		   states ,  "STATE"    ,  0 ,     calls  ,  "CALL"   ,  0 ,
		   bye	  ,  "BYE"      ,  0 ,     words  ,  "WORDS"  ,  0 ,
		   dots	  ,  ".S"      ,  0 ,   
		   literal,  "LITERAL"  ,  1 ,     incl   ,  "INCLUDE",  0 ,
		   word   ,  "WORD"     ,  0 ,     ms     ,  "MSEC"   ,  0 ,
		   comp   ,  "]"        ,  0 ,     forget ,  "FORGET" ,  0 ,
                   execute,  "EXECUTE"  ,  0 ,     kill   ,  "kill"   ,  0 ,
                   forth  ,  "FORTH"    ,  0 ,     comment,  "("      ,  1 ,
                   colon  ,  ":"        ,  0 };
/*
The following Forth compiles at program startup.
-------------------------------------------------------------------------------
The entertainment value of seeing Forth in the middle of a C program probably
justifies the ram used by this string. If you wish to save the ram, move these
definitions to a text file and
*/
char *startup =

"  : ABORT    kill ;"
"  : ,        HERE !  1 ALLOT              ;           "
"  : THEN     HERE SWAP !                  ; IMMEDIATE "
"  : IF       | if HERE  0 ,               ; IMMEDIATE "
"  : ELSE     | branch HERE  0 , SWAP | THEN ; IMMEDIATE "
"  : BEGIN    HERE                         ; IMMEDIATE "
"  : AGAIN    | branch ,                   ; IMMEDIATE "
"  : WHILE    | IF  SWAP                   ; IMMEDIATE "
"  : REPEAT   | AGAIN | THEN               ; IMMEDIATE "
"  : UNTIL    | if ,                       ; IMMEDIATE "
"  : RECURSE  LAST CALL                    ; IMMEDIATE "
"  : DO       | SWAP | >R | >R  HERE       ; IMMEDIATE "
"  : undo     R> R> DROP R> DROP >R        ;           "
"  : UNDO     | undo                       ; IMMEDIATE " // Like Will Baden's
"  : LOOP     |  loop ,                    ; IMMEDIATE "
"  : +LOOP    | +loop ,                    ; IMMEDIATE "
"  : DOES     R> LAST AT 1+ !              ;           "
"  : DOES>    | DOES | R>                  ; IMMEDIATE "
"  : CREATE   HEAD  | nest 0 ,  DOES>      ;           "
"  : COUNT    DUP 1+ SWAP @                ;           "
"  : ?DUP     DUP IF DUP THEN              ;           "
"  : TYPE     ?DUP IF 0 DO COUNT EMIT LOOP THEN DROP ;     "
"  : quote    R> COUNT OVER OVER + >R TYPE ;               "
"  : .\"      | quote  34 WORD @ 1+ ALLOT ; IMMEDIATE      "
"  : string   R> DUP COUNT + >R ;                          "
"  : \"       | string 34 WORD @ 1+ ALLOT ; IMMEDIATE      "
"  : VALUE    HEAD | LITERAL | EXIT      ;                 "
"  : TO       ' AT 1+  STATE                               "
"             IF | LITERAL | !  ELSE ! THEN ; IMMEDIATE    "
"  : CONSTANT CREATE IMMEDIATE ,  DOES> @  STATE IF | LITERAL THEN ; "
"  : VARIABLE HERE 3 + CONSTANT   0 , ;  0 CONSTANT BASE "
"  : DECIMAL  10 BASE ! ; "
"  : HEX      16 BASE ! ; "
"  : SPACE    32 EMIT   ; "
"  : HOLD     SWAP >R SWAP R> ; "
"  : <#       0 HOLD ;    "
"  : #>       DROP DROP BEGIN DUP WHILE EMIT REPEAT DROP ;  "
"  : DIGIT    9 OVER < 7 AND + 48 + ;                       "
"  : #        0 BASE @ UM/MOD >R BASE @ UM/MOD R> ROT DIGIT HOLD ; "
"  : #S       BEGIN # OVER OVER OR WHILE REPEAT ;           "
"  : UD.      <# #S #> SPACE ;                              "
"  : D.       DUP 0< IF 45 EMIT DNEGATE THEN UD. ;          "
"  : U.       0 UD. ; "
"  : .        DUP 0< D. ; "
"  : INT      1 2 ! ; "
"  : FP       0 2 ! ; "
// 
"  : VIDEO    1 1 ! ; "
"  : SQUARE   0 1 ! ; "
"  : ->       ' AT 2 + @ | astore , ; IMMEDIATE"
"  : PIXEL    HEAD IMMEDIATE 256 * +  HERE 1+ @ 78 = IF 65536 + THEN "
"             | nest 0 , , DOES> | afetch @ , ; " ;
/* 
Interpreter and compiler
------------------------------------------------------------------------------*/
int  wort(char d) { int * hh=h; *hh++=0;	     /* Initialize HERE       */
	  char *w = SAID-1;			     /* Takes ascii delimiter.*/
	  while (*in == d) in++;		     /* Ignor leading delims. */
	  while (((*hh++ = *++w = *in++) != d) && *w)/* Copy word to SAID and */
		(*h)++; 			     /* HERE till delim or end*/
	  if (!*w) in--;			     /* and backtrack if end. */
	  *w = 0;				     /* null-terminate	      */
	  return w == SAID; }			     /* return end-of-input   */
int  same(char *a, char *b) {			     /* String compare	      */
	  while(*a)  if (*a++ != *b++)	return 0;
	  return !*b; }
int  find(char *s)  { op = top+state;			  /* Return index if  */
	  while (!same(s,heads[op].name)) if(!--op)break; /* in dictionary or */
	  return op; }					  /* 0. 	      */

int  get()          { while(wort(32)) {
                        if (!input) return 0; record(); } return 1; }
void comment()      { if (!input) wort(41);
                      else { while(input && *in++ != 41) if(!*in) record(); } }
int  addr(int n)    { return heads[n].code.ram; }
void warn(char *c)  { s=stack;	r=rs; t=0; sout(c); sout("?"); in=tib; *in=0;
		      if (input) { CloseHandle(input);
				   input = 0; } }
void call(int n)    { if (n>ctop) { *h++ = heads[0].code.ram;
                                    *h++ = addr(n); }
                      else          *h++ = heads[n].code.ram; }
void number() { char * end = SAID;
	  PUSH(0);
          if ( ram[2]) t = strtol(SAID, &end, ram[0]);
          if (*end == ' ') return ;
          if (!ram[2] || *end == '.' || *end == '-') {
		     f = strtod(SAID, &end ); }
          if (*end) warn(SAID); }
void compile() { state = -1;			     /* Set compiling	     */
	  while (get()) {			     /* Stop if string end   */
	     if (find(SAID)) {
                    if (heads[op].immediate) {
                       PUSH(op); execute(); }        /* Execute an immediate */
		    else call(op);		     /* Compile if found     */
		    if (op==1) break; } 	     /* Stop if ;	     */
	     else { number();
		    if (in==tib) {
		       h=heads[top--].code.ram+ram;  /* Forget bad definition*/
		       break; }
		    literal(); } }		     /* Compile number	     */
	  state = 0; }
void interpret() {
	  while (get()) {
             if (find(SAID)) { PUSH(op); execute(); }
	     else number();
	     if (s<stack | r<rs)  warn("Stack"); } }
//
int WINAPI WinMain(HINSTANCE This, HINSTANCE Prev, LPSTR Command, int mode) {
  MSG	     msg;
  WNDCLASSEX c;
  c.cbSize	  = sizeof(WNDCLASSEX);
  c.hInstance	  = This;			     // Handle of application
  c.lpszClassName = "NewConsole";                    // Window class name
  c.lpfnWndProc   = wfunc;			     // Window function
  c.style	  = 0;				     // Default style
  c.hIcon	  = LoadIcon(NULL, IDI_APPLICATION); // Standard icon
  c.hIconSm	  = LoadIcon(NULL, IDI_APPLICATION); // Small icon
  c.hCursor	  = LoadCursor(NULL, IDC_ARROW);     // Cursor style
  c.lpszMenuName  = NULL;			     // No menu
  c.cbClsExtra	  = 0;				     // No extra info
  c.cbWndExtra	  = 0;				     // No extra info
  c.hbrBackground = (HBRUSH)
		    GetStockObject(WHITE_BRUSH);     // White background
  if(!RegisterClassEx(&c)) return 0;		     // Register class
  window = CreateWindow( "NewConsole",               // Window class name
                         "Win32 riForth  Copyright Robert F. Illyes, 1994-2000",
			 WS_OVERLAPPEDWINDOW,	     // Normal style
			 CW_USEDEFAULT, 	     // X position
			 CW_USEDEFAULT, 	     // Y position
			 wide,			     // Width
			 high,			     // Neight
			 HWND_DESKTOP,		     // No parent
			 NULL,			     // no menu
			 This,			     // Handle of application
			 NULL );		     // Not MDI
  ShowWindow(window, mode);
  HDC screen = BeginPaint(window, &paint);
  bitmap     = CreateCompatibleDC(screen);
  BITMAPINFOHEADER hinfo = { sizeof(BITMAPINFOHEADER), 649, 433, 1, 24, BI_RGB};
  BITMAPINFO info; info.bmiHeader = hinfo;
  bits       = CreateDIBSection(screen, &info, DIB_RGB_COLORS, (void **)&pix, 0, 0);
  SelectObject(bitmap, bits);
  color      = RGB(0, 0, 255);
  pen	     = CreatePen(PS_SOLID, 1, color);
  SelectObject(bitmap, pen);
  SelectObject(bitmap, GetStockObject(OEM_FIXED_FONT));
  EndPaint(window, &paint);
  clear();
  el = 1;
  heads= ops;
  ctop = top = find(":");
  ram[0] = 10; // Initialize BASE to DECIMAL
  ram[1] =  0; // Initialize video flag to SQUARE
  ram[2] =  1; // Initialize to mixed integer/floating numeric input
  in = startup;
  interpret();
  if (*Command) { in = Command; get(); opener(h - ram); interpret(); }
  in = tib;
  emt('_');
  force();
  while(GetMessage(&msg, NULL, 0, 0)) {
      TranslateMessage(&msg);
      DispatchMessage(&msg); }
  return msg.wParam;
}
//
/* Stuff to do:

1) Bring more compiler words out into the dictionary.
2) Include the entire graphics tool box.
3) INCLUDE nesting.
4) Branch termination.
5) Unsigned integer input that includes the sign bit.
6) File output, creation, COPY, PRINT, CAT, and LS.
7) Printer output with graphics support.
  */
